import os
import pytest
import wandb
from wandb import artifacts
from wandb import util
from .utils import runner

def mock_boto(artifact, path=False):
    class S3Object(object):
        def __init__(self, name='my_object.pb', metadata=None):
            self.metadata = metadata or {"md5": "1234567890abcde"}
            self.e_tag = '"1234567890abcde"'
            self.version_id = "1"
            self.name = name
            self.key = name
            self.content_length = 10

        def load(self):
            if path:
                raise util.get_module("botocore").exceptions.ClientError({"Error": {"Code": "404"}}, "HeadObject")

    class Filtered(object):
        def limit(self, *args, **kwargs):
            return [S3Object(), S3Object(name="my_other_object.pb")]

    class S3Objects(object):
        def filter(self, **kwargs):
            return Filtered()

    class S3Bucket(object):
        def __init__(self, *args, **kwargs):
            self.objects = S3Objects()

    class S3Resource(object):
        def Object(self, bucket, key):
            return S3Object()

        def Bucket(self, bucket):
            return S3Bucket()

    mock = S3Resource()
    handler = artifact._storage_policy._handler._handlers["s3"]
    handler._s3 = mock
    handler._botocore = util.get_module("botocore")
    return mock

def mock_gcs(artifact, path=False):
    class Blob(object):
        def __init__(self, name='my_object.pb', metadata=None):
            self.md5_hash = "1234567890abcde"
            self.etag = '1234567890abcde'
            self.generation = "1"
            self.name = name
            self.size = 10

    class GSBucket(object):
        def get_blob(self,*args, **kwargs):
            return None if path else Blob()

        def list_blobs(self, *args, **kwargs):
            return [Blob(), Blob(name="my_other_object.pb")]

    class GSClient(object):
        def bucket(self, bucket):
            return GSBucket()

    mock = GSClient()
    handler = artifact._storage_policy._handler._handlers["gs"]
    handler._client = mock
    return mock

def test_add_one_file(runner):
    with runner.isolated_filesystem():
        with open('file1.txt', 'w') as f:
            f.write('hello')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_file('file1.txt')

        assert artifact.digest == 'a00c2239f036fb656c1dcbf9a32d89b4'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['file1.txt'] == {
            'digest': 'XUFAKrxLKna5cZ2REBfFkg==', 'size': 5}

def test_add_named_file(runner):
    with runner.isolated_filesystem():
        with open('file1.txt', 'w') as f:
            f.write('hello')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_file('file1.txt', name='great-file.txt')

        assert artifact.digest == '585b9ada17797e37c9cbab391e69b8c5'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['great-file.txt'] == {
            'digest': 'XUFAKrxLKna5cZ2REBfFkg==', 'size': 5}

def test_add_new_file(runner):
    with runner.isolated_filesystem():
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        with artifact.new_file('file1.txt') as f:
            f.write('hello')

        assert artifact.digest == 'a00c2239f036fb656c1dcbf9a32d89b4'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['file1.txt'] == {
            'digest': 'XUFAKrxLKna5cZ2REBfFkg==', 'size': 5}

def test_add_dir(runner):
    with runner.isolated_filesystem():
        open('file1.txt', 'w').write('hello')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_dir('.')

        assert artifact.digest == 'a00c2239f036fb656c1dcbf9a32d89b4'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['file1.txt'] == {
            'digest': 'XUFAKrxLKna5cZ2REBfFkg==', 'size': 5}

def test_add_named_dir(runner):
    with runner.isolated_filesystem():
        open('file1.txt', 'w').write('hello')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_dir('.', name='subdir')

        assert artifact.digest == 'a757208d042e8627b2970d72a71bed5b'

        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['subdir/file1.txt'] == {
            'digest': 'XUFAKrxLKna5cZ2REBfFkg==', 'size': 5}

def test_add_reference_local_file(runner):
    with runner.isolated_filesystem():
        open('file1.txt', 'w').write('hello')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_reference('file://file1.txt')

        assert artifact.digest == 'a00c2239f036fb656c1dcbf9a32d89b4'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['file1.txt'] == {
            'digest': 'XUFAKrxLKna5cZ2REBfFkg==', 'ref': 'file://file1.txt', 'size': 5}

def test_add_reference_local_file_no_checksum(runner):
    with runner.isolated_filesystem():
        open('file1.txt', 'w').write('hello')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_reference('file://file1.txt', checksum=False)

        assert artifact.digest == '2f66dd01e5aea4af52445f7602fe88a0'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['file1.txt'] == {
            'digest': 'file://file1.txt', 'ref': 'file://file1.txt'}

def test_add_reference_local_dir(runner):
    with runner.isolated_filesystem():
        open('file1.txt', 'w').write('hello')
        open('file2.txt', 'w').write('dude')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_reference('file://'+os.getcwd())

        assert artifact.digest == '5e8e98ebd59cc93b58d0cb26432d4720'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['file1.txt'] == {
            'digest': 'XUFAKrxLKna5cZ2REBfFkg==', 'ref': 'file://'+os.getcwd()+'/file1.txt', 'size': 5}
        assert manifest['contents']['file2.txt'] == {
            'digest': 'E7c+2uhEOZC+GqjxpIO8Jw==', 'ref': 'file://'+os.getcwd()+'/file2.txt', 'size': 4}

def test_add_s3_reference_object(runner, mocker):
    with runner.isolated_filesystem():
        artifact = artifacts.Artifact(type="dataset", name='my-arty')
        mock_boto(artifact)
        artifact.add_reference("s3://my-bucket/my_object.pb")

        assert artifact.digest == '8aec0d6978da8c2b0bf5662b3fd043a4'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['my_object.pb'] == {
            'digest': '1234567890abcde', 'ref': 's3://my-bucket/my_object.pb',
            'extra': {'etag': '1234567890abcde', 'versionID': '1'}, 'size': 10}

def test_add_s3_reference_path(runner, mocker, capsys):
    with runner.isolated_filesystem():
        artifact = artifacts.Artifact(type="dataset", name='my-arty')
        mock_boto(artifact, path=True)
        artifact.add_reference("s3://my-bucket/")

        assert artifact.digest == '17955d00a20e1074c3bc96c74b724bfe'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['my_object.pb'] == {
            'digest': '1234567890abcde', 'ref': 's3://my-bucket/my_object.pb',
            'extra': {'etag': '1234567890abcde', 'versionID': '1'}, 'size': 10}
        _, err = capsys.readouterr()
        assert "Generating checksum" in err

def test_add_s3_max_objects(runner, mocker, capsys):
    with runner.isolated_filesystem():
        artifact = artifacts.Artifact(type="dataset", name='my-arty')
        mock_boto(artifact, path=True)
        with pytest.raises(ValueError):
            artifact.add_reference("s3://my-bucket/", max_objects=1)

def test_add_reference_s3_no_checksum(runner):
    with runner.isolated_filesystem():
        open('file1.txt', 'w').write('hello')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        # TODO: Should we require name in this case?
        artifact.add_reference('s3://my_bucket/file1.txt', checksum=False)

        assert artifact.digest == '52631787ed3579325f985dc0f2374040'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['file1.txt'] == {
            'digest': 's3://my_bucket/file1.txt', 'ref': 's3://my_bucket/file1.txt'}

def test_add_gs_reference_object(runner, mocker):
    with runner.isolated_filesystem():
        artifact = artifacts.Artifact(type="dataset", name='my-arty')
        mock_gcs(artifact)
        artifact.add_reference("gs://my-bucket/my_object.pb")

        assert artifact.digest == '8aec0d6978da8c2b0bf5662b3fd043a4'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['my_object.pb'] == {
            'digest': '1234567890abcde', 'ref': 'gs://my-bucket/my_object.pb',
            'extra': {'etag': '1234567890abcde', 'versionID': '1'}, 'size': 10}

def test_add_gs_reference_path(runner, mocker, capsys):
    with runner.isolated_filesystem():
        artifact = artifacts.Artifact(type="dataset", name='my-arty')
        mock_gcs(artifact, path=True)
        artifact.add_reference("gs://my-bucket/")

        assert artifact.digest == '17955d00a20e1074c3bc96c74b724bfe'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['my_object.pb'] == {
            'digest': '1234567890abcde', 'ref': 'gs://my-bucket/my_object.pb',
            'extra': {'etag': '1234567890abcde', 'versionID': '1'}, 'size': 10}
        _, err = capsys.readouterr()
        assert "Generating checksum" in err

def test_add_reference_named_local_file(runner):
    with runner.isolated_filesystem():
        open('file1.txt', 'w').write('hello')
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_reference('file://file1.txt', name='great-file.txt')

        assert artifact.digest == '585b9ada17797e37c9cbab391e69b8c5'
        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['great-file.txt'] == {
            'digest': 'XUFAKrxLKna5cZ2REBfFkg==', 'ref': 'file://file1.txt', 'size': 5}

def test_add_reference_unknown_handler(runner):
    with runner.isolated_filesystem():
        artifact = artifacts.Artifact(type='dataset', name='my-arty')
        artifact.add_reference('http://example.com/somefile.txt', name='ref')

        assert artifact.digest == '5b8876252f3ca922c164de380089c9ae'

        manifest = artifact.manifest.to_manifest_json()
        assert manifest['contents']['ref'] == {
            'digest': 'http://example.com/somefile.txt', 'ref': 'http://example.com/somefile.txt'}

def test_log_artifact_simple(runner, wandb_init_run):
    util.mkdir_exists_ok("artsy")
    open("artsy/file1.txt", "w").write("hello")
    open("artsy/file2.txt", "w").write("goodbye")
    with pytest.raises(ValueError):
        wandb.log_artifact("artsy")
    art = wandb.log_artifact("artsy", type="dataset")
    assert art.name == "run-"+wandb_init_run.id+"-artsy"

def test_use_artifact_simple(runner, wandb_init_run):
    art = wandb.use_artifact("mnist:v0", type="dataset")
    assert art.name == "mnist:v0"
    path = art.download()
    assert os.path.exists(path)


import wandb
from sklearn.datasets import make_regression
from catboost import CatBoostRegressor, CatBoostClassifier


def test_types(**kwargs):
    test_passed = True
    for k, v in kwargs.items():
        if k == 'model':
            if not (isinstance(v, CatBoostRegressor) or isinstance(v, CatBoostClassifier)):
                test_passed = False
                wandb.termerror("%s is not a regressor or classifier. Please try again." % (k))

    return test_passed


def test_fitted(model):
    try:
        X, y = make_regression(n_features=2, random_state=24)
        temp_model = model.copy()
        temp_model.fit(X, y, verbose=False)
    except Exception:
        wandb.termerror("Please fit the model before passing it in.")
        return False

    return True



# This Python file uses the following encoding: utf-8
import pytest
import datetime
import os
import collections
import sys
import json
from .utils import git_repo

from wandb import wandb_run
from wandb import env
from wandb.apis import InternalApi, CommError

# Tests which rely on row history in memory should set `History.keep_rows = True`
from wandb.history import History
History.keep_rows = True


def get_last_val(history, key):
    val = None
    for val in history.column(key):
        pass
    return val

def test_wandb_run_args(git_repo):
    environ = dict(os.environ)
    environ[env.ARGS] = json.dumps(["foo", "bar"])
    run = wandb_run.Run.from_environment_or_defaults(environ)
    assert run.args == ["foo", "bar"]


def test_url_escape(git_repo):
    environ = dict(os.environ)
    environ[env.ENTITY] = "†est"
    environ[env.PROJECT] = "wild projo"
    environ[env.API_KEY] = "abcdefghijabcdefghijabcdefghijabcdefghij"
    environ[env.RUN_ID] = "my wild run"
    run = wandb_run.Run.from_environment_or_defaults(environ)
    assert run.get_url() == 'https://app.wandb.ai/%E2%80%A0est/wild%20projo/runs/my%20wild%20run'


def test_url_escape_query_string(git_repo):
    environ = dict(os.environ)
    environ[env.ENTITY] = "†est"
    environ[env.PROJECT] = "wild projo"
    environ[env.API_KEY] = "abcdefghijabcdefghijabcdefghijabcdefghij"
    environ[env.RUN_ID] = "my wild run"
    run = wandb_run.Run.from_environment_or_defaults(environ)
    assert run.get_url(params=collections.OrderedDict([
        ('first', 'abc123'),
        ('second', ' '),
    ])) == 'https://app.wandb.ai/%E2%80%A0est/wild%20projo/runs/my%20wild%20run?first=abc123&second=+'


def test_wandb_run_args_sys(git_repo, mocker):
    environ = dict(os.environ)
    if env.ARGS in environ:
        del environ[env.ARGS]  # force our code to use sys.argv.
    mocker.patch.object(sys, 'argv', ["rad", "cool"])
    run = wandb_run.Run.from_environment_or_defaults(environ)
    assert run.args == ["cool"]


def test_name_and_desc_only_name(git_repo):
    environ = dict(os.environ)
    if env.NAME in environ:
        del environ[env.NAME]
    if env.NOTES in environ:
        del environ[env.NOTES]
    environ[env.DESCRIPTION] = "myrunid"
    run = wandb_run.Run.from_environment_or_defaults(environ)
    assert run.name == "myrunid"
    assert run.description == ""


def test_name_and_desc(git_repo):
    environ = dict(os.environ)
    if env.NAME in environ:
        del environ[env.NAME]
    if env.NOTES in environ:
        del environ[env.NOTES]
    environ[env.DESCRIPTION] = "myrunid\nmydesc"
    run = wandb_run.Run.from_environment_or_defaults(environ)
    assert run.name == "myrunid"
    assert run.description == "mydesc"


def test_name_and_desc_setters(git_repo):
    run = wandb_run.Run.from_environment_or_defaults()
    run.name = "123"
    run.description = "so much desc\nthis is fun"
    assert run.name == "123"
    assert run.description == "so much desc\nthis is fun"
    my_env = {}
    run.set_environment(my_env)
    assert my_env[env.DESCRIPTION] == "123\nso much desc\nthis is fun"


def test_get_url(git_repo, loggedin):
    api = InternalApi({"entity": "cool"})
    api.set_setting("anonymous", "true")
    run = wandb_run.Run(api=api)
    assert run.get_url() == "https://app.wandb.ai/cool/uncategorized/runs/"+run.id+"?apiKey="+"X"*40
    assert run.get_project_url() == "https://app.wandb.ai/cool/uncategorized?apiKey="+"X"*40
    api.set_setting("entity", "")
    with pytest.raises(CommError):
        run.get_url()

def test_set_setting_no_persist_by_default(git_repo):
    os.remove("wandb/settings")
    api = InternalApi({"entity": "cool"})
    api.set_setting("rad", "true")
    assert not os.path.exists("wandb/settings")
    assert api.settings("rad") == "true"


def test_history_updates_keys_until_summary_writes(git_repo):
    run = wandb_run.Run()

    run.history.add({'a': 5, 'b': 9})
    assert get_last_val(run.history, 'a') == 5
    assert get_last_val(run.history, 'b') == 9
    assert run.summary['a'] == 5
    assert run.summary['b'] == 9

    run.history.add({'a': 6, 'b': 10})
    assert get_last_val(run.history, 'a') == 6
    assert get_last_val(run.history, 'b') == 10
    assert run.summary['a'] == 6
    assert run.summary['b'] == 10

    run.summary['a'] = 112491
    assert run.summary['a'] == 112491

    run.history.add({'a': 1, 'b': 3})
    assert get_last_val(run.history, 'a') == 1
    assert get_last_val(run.history, 'b') == 3
    assert run.summary['a'] == 112491
    assert run.summary['b'] == 3

    run.history.add({'a': -40, 'b': -49})
    assert get_last_val(run.history, 'a') == -40
    assert get_last_val(run.history, 'b') == -49
    assert run.summary['a'] == 112491
    # most recent history key is logged
    assert run.summary['b'] == -49

    run.summary['c'] = 100
    assert run.summary['c'] == 100
    run.history.add({'c': 200, 'd': 300})
    assert run.summary['c'] == 100
    assert run.summary['d'] == 300

    run.history.add({'a': 1000, 'b': 2000, 'c': 200, 'd': 300})
    assert get_last_val(run.history, 'a') == 1000
    assert get_last_val(run.history, 'b') == 2000
    assert get_last_val(run.history, 'c') == 200
    assert get_last_val(run.history, 'd') == 300
    assert run.summary['a'] == 112491
    assert run.summary['b'] == 2000
    assert run.summary['c'] == 100
    assert run.summary['d'] == 300



"""Mock Server for simple calls the cli and public api makes"""

from flask import Flask, request
import os
from datetime import datetime
import json

def run():
    return {
        'id': 'test',
        'name': 'wild-test',
        'displayName': 'beast-bug-33',
        'state': "running",
        'config': '{"epochs": {"value": 10}}',
        'description': "",
        'systemMetrics': '{"cpu": 100}',
        'summaryMetrics': '{"acc": 100, "loss": 0}',
        'fileCount': 1,
        'history': [
            '{"acc": 10, "loss": 90}',
            '{"acc": 20, "loss": 80}',
            '{"acc": 30, "loss": 70}'
        ],
        'events': [
            '{"cpu": 10}',
            '{"cpu": 20}',
            '{"cpu": 30}'
        ],
        "files": {
            # Special weights url meant to be used with api_mocks#download_url
            "edges": [{"node": {"name": "weights.h5", "sizeBytes": 20, "url": request.url_root + "/storage?file=weights.h5"}}]
        },
        'tags': [],
        'notes': None,
        'sweepName': None,
    }

def artifact(ctx, collection_name="mnist"):
    return {
        "id": ctx["page_count"],
        "digest": "abc123",
        "description": "",
        "state": "COMMITTED",
        "size": 10000,
        "createdAt": datetime.now().isoformat(),
        "updatedAt": datetime.now().isoformat(),
        "labels": [],
        "metadata": "{}",
        "aliases": [{
            "artifactCollectionName": collection_name,
            "alias": "v%i" % ctx["page_count"]
        }]
    }

def paginated(node, ctx, extra={}):
    next_page = False
    ctx["page_count"] += 1
    if ctx["page_count"] < ctx["page_times"]:
        next_page = True
    edge = {
        "node": node,
        "cursor": "abc123"
    }
    edge.update(extra)
    return {
        "edges": [edge],
        "pageInfo": {
            "endCursor": "abc123",
            "hasNextPage": next_page
        }
    }


def create_app(ctx):
    app = Flask(__name__)

    @app.route("/graphql", methods=["POST"])
    def graphql():
        if "fail_times" in ctx:
            if ctx["fail_count"] < ctx["fail_times"]:
                ctx["fail_count"] += 1
                return json.dumps({"errors": ["Server down"]}), 500
        body = request.get_json()
        if body["variables"].get("files"):
            file = body["variables"]["files"][0]
            return json.dumps({
                "data": {
                    "model": {
                        "bucket": {
                            "id": "storageid",
                            "files": {
                                "uploadHeaders": [],
                                "edges": [{"node": {"name": file, "url": request.url_root + "/storage?file=%s" % file}}]
                            }
                        }
                    }
                }
            })
        if "historyTail" in body["query"]:
            return json.dumps({
                'data': {
                    'model': {
                        'bucket': {
                            'name': "test",
                            'displayName': 'funky-town-13',
                            'id': "test",
                            'summaryMetrics': '{"acc": 10}',
                            'logLineCount': 14,
                            'historyLineCount': 15,
                            'eventsLineCount': 0,
                            'historyTail': '["{\\"_step\\": 15, \\"acc\\": 1}"]',
                            'eventsTail': '[]'
                        }
                    }
                }
            })
        if "query Runs(" in body["query"]:
            return json.dumps({
                "data": {
                    "project": {
                        "runCount": 4,
                        "readOnly": False,
                        "runs": paginated(run(), ctx)
                    }
                }
            })
        if "query Run(" in body["query"]:
            return json.dumps({
                'data': {
                    'project': {
                        'run': run()
                    }
                }
            })
        if "query Projects(" in body["query"]:
            return json.dumps({
                "data": {
                    "models": paginated({
                        "id": "1",
                        "name": "test-project",
                        "entityName": body["variables"]["entity"],
                        "createdAt": "now",
                        "isBenchmark": False,
                    }, ctx)
                }
            })
        if "query Viewer " in body["query"]:
            return json.dumps({
                "data": {
                    "viewer": {
                        "entity": "vanpelt",
                        "flags": '{"code_saving_enabled": true}'
                    }
                }
            })
        if "mutation UpsertBucket(" in body["query"]:
            return json.dumps({
                "data": {
                    "upsertBucket": {
                        "bucket": {
                            "id": "storageid",
                            "name": body["variables"].get("name", "abc123"),
                            "displayName": "lovely-dawn-32",
                            "project": {
                                "name": "test",
                                "entity": {
                                    "name": "mock_server_entity"
                                }
                            }
                        }
                    }
                }
            })
        if "mutation CreateAnonymousApiKey " in body["query"]:
            return json.dumps({
                "data": {
                    "createAnonymousEntity": {
                        "apiKey": {
                            "name": "ANONYMOOSE" * 4
                        }
                    }
                }
            })
        if "mutation PrepareFiles(" in body["query"]:
            nodes = []
            for i, file_spec in enumerate(body['variables']['fileSpecs']):
                nodes.append({
                    "node" : {
                        "id": str(i),
                        "name": file_spec['name'],
                        "displayName": file_spec['name'],
                        "digest": "null",
                        "uploadUrl": request.url_root + "/storage?file=%s" % file_spec['name'],
                        "uploadHeaders": ""
                    }})
            return json.dumps({
                "data": {
                    "prepareFiles": {
                        "files": {
                            "edges": nodes
                        }
                    }
                }
            })
        if "mutation CreateArtifact(" in body["query"]:
            return {
                "data": {
                    "createArtifact": {
                        "artifact": artifact(ctx, body["variables"]["artifactCollectionNames"][0])
                    }
                }
            }
        if "mutation UseArtifact(" in body["query"]:
            return {
                "data": {
                    "useArtifact": {
                        "artifact": artifact(ctx)
                    }
                }
            }
        if "query ProjectArtifactType(" in body["query"]:
            return {
                "data": {
                    "project": {
                        "artifactType": {
                            "id": "1",
                            "name": "dataset",
                            "description": "",
                            "createdAt": datetime.now().isoformat()
                        }
                    }
                }
            }
        if "query ProjectArtifacts(" in body["query"]:
            return {
                "data": {
                    "project": {
                        "artifactTypes": paginated({
                            "id": "1",
                            "name": "dataset",
                            "description": "",
                            "createdAt": datetime.now().isoformat()
                        }, ctx)
                    }
                }
            }
        if "query ProjectArtifactCollections(" in body["query"]:
            return {
                "data": {
                    "project": {
                        "artifactType": {
                            "artifactSequences": paginated({
                                "id": "1",
                                "name": "mnist",
                                "description": "",
                                "createdAt": datetime.now().isoformat()
                            }, ctx)
                        }
                    }
                }
            }
        if "query RunArtifacts(" in body["query"]:
            key = "inputArtifacts" if "inputArtifacts" in body["query"] else "outputArtifacts"
            artifacts = paginated(artifact(ctx), ctx)
            artifacts["totalCount"] = ctx["page_times"]
            return {
                "data": {
                    "project": {
                        "run": {
                            key: artifacts
                        }
                    }
                }
            }
        if "query Artifacts(" in body["query"]:
            artifacts = paginated(artifact(ctx), ctx, {"version": "v%i" % ctx["page_count"]})
            artifacts["totalCount"] = ctx["page_times"]
            return {
                "data": {
                    "project": {
                        "artifactType": {
                            "artifactSequence": {
                                "name": "mnist",
                                "artifacts": artifacts
                            }
                        }
                    }
                }
            }
        if "query Artifact(" in body["query"]:
            art = artifact(ctx)
            art["artifactType"] = {
                "id": 1,
                "name": "dataset"
            }
            art["currentManifest"] = {
                "id": 1,
                "file": {
                    "id": 1,
                    "url": request.url_root + "/storage?file=wandb_manifest.json"
                }
            }
            return {
                "data": {
                    "project": {
                        "artifact": art
                    }
                }
            }
        if "stopped" in body["query"]:
            return json.dumps({
                "data": {
                    "Model": {
                        "project": {
                            "run": {
                                "stopped": False
                            }
                        }
                    }
                }
            })
        print("MISSING QUERY", body["query"])
        return json.dumps({"errors": [{"message": "Not implemented in tests/mock_server.py", "body": body}]})

    @app.route("/storage", methods=["PUT", "GET"])
    def storage():
        file = request.args.get('file')
        size = ctx["files"].get(request.args.get('file'))
        if request.method == "GET" and size:
            return os.urandom(size), 200
        if file == "wandb_manifest.json":
            return {
                "version": 1,
                "storagePolicy": "wandb-storage-policy-v1",
                "storagePolicyConfig": {},
                "contents": {
                    "digits.h5": {
                        "digest": "TeSJ4xxXg0ohuL5xEdq2Ew==",
                        "size": 81299
                    }
                }}
        return "", 200

    @app.route("/artifacts/<entity>/<digest>", methods=["GET", "POST"])
    def artifact_file(entity, digest):
        return "ARTIFACT %s" % digest, 200

    @app.route("/files/<entity>/<project>/<run>/file_stream", methods=["POST"])
    def file_stream(entity, project, run):
        return json.dumps({"exitcode": None, "limits": {}})

    @app.route("/api/v1/namespaces/default/pods/test")
    def k8s_pod():
        if ctx.get("k8s"):
            return b'{"status":{"containerStatuses":[{"imageID":"docker-pullable://test@sha256:1234"}]}}', 200
        else:
            return b'', 500

    @app.errorhandler(404)
    def page_not_found(e):
        print("Got request to: %s (%s)" % (request.url, request.method))
        return "Not Found", 404

    return app


if __name__== '__main__':
    app = create_app()



from .utils import runner
import os
import sh
import sys
import glob
import json
import time
import signal
import tempfile
import pytest

train_py = open(os.path.join(os.path.dirname(
    __file__), "fixtures/train.py")).read()


def test_dry_run(runner):
    with runner.isolated_filesystem():
        with open("train.py", "w") as f:
            f.write(train_py)
        environ = {
            "WANDB_MODE": "dryrun",
            "WANDB_TEST": "true",
        }
        res = sh.python("train.py", _env=environ)
        run_dir = glob.glob(os.path.join("wandb", "dry*"))[0]
        meta = json.loads(
            open(os.path.join(run_dir, "wandb-metadata.json")).read())
        assert meta["state"] == "finished"
        assert meta["program"] == "train.py"
        assert meta["exitcode"] == 0
        assert os.path.exists(os.path.join(run_dir, "output.log"))
        # TODO: FLAKY SPEC Sometimes output.log is empty...
        assert "loss:" in open(os.path.join(run_dir, "output.log")).read()
        assert os.path.exists(os.path.join(
            run_dir, "wandb-history.jsonl"))
        assert os.path.exists(os.path.join(run_dir, "wandb-events.jsonl"))
        assert os.path.exists(os.path.join(run_dir, "wandb-summary.json"))


def test_dry_run_custom_dir(runner):
    with runner.isolated_filesystem():
        environ = {
            "WANDB_DIR": tempfile.mkdtemp(),
            "WANDB_MODE": "dryrun",
            "WANDB_TEST": "true",
        }
        try:
            with open("train.py", "w") as f:
                f.write(train_py)
            res = sh.python("train.py", _env=environ, _err=1)
            print(res)
            run_dir = glob.glob(os.path.join(
                environ["WANDB_DIR"], "wandb", "dry*"))[0]
            assert os.path.exists(run_dir + "/output.log")
        finally:  # avoid stepping on other tests, even if this one fails
            sh.rm("-rf", environ["WANDB_DIR"])


def test_dry_run_exc(runner):
    with runner.isolated_filesystem():
        with open("train.py", "w") as f:
            f.write(train_py.replace("#raise", "raise"))
        environ = {
            "WANDB_MODE": "dryrun",
            "WANDB_TEST": "true",
        }
        try:
            res = sh.python("train.py", _env=environ)
        except sh.ErrorReturnCode as e:
            res = e.stderr
        print(res)
        run_dir = glob.glob(os.path.join("wandb", "dry*"))[0]
        meta = json.loads(
            open(os.path.join(run_dir, "wandb-metadata.json")).read())
        assert meta["state"] == "failed"
        assert meta["exitcode"] == 1


def test_dry_run_kill(runner):
    with runner.isolated_filesystem():
        with open("train.py", "w") as f:
            f.write(train_py.replace("#os.kill", "os.kill"))
        environ = {
            "WANDB_MODE": "dryrun",
            "WANDB_TEST": "true",
        }
        res = sh.python("train.py", epochs=10, _bg=True, _env=environ)
        try:
            res.wait()
            print(res)
        except sh.ErrorReturnCode:
            pass
        dirs = glob.glob(os.path.join("wandb", "dry*"))
        print(dirs)
        run_dir = dirs[0]
        meta = json.loads(
            open(os.path.join(run_dir, "wandb-metadata.json")).read())
        assert meta["state"] == "killed"
        assert meta["exitcode"] == 255
        assert meta["args"] == ["--epochs=10"]


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Funky things in python 2 land and multiprocessing")
def test_mock_server_no_internet(runner):
    with runner.isolated_filesystem():
        with open("train.py", "w") as f:
            f.write(train_py)
        environ = {
            "WANDB_BASE_URL": "http://localhost:%i" % 10101,
            "WANDB_API_KEY": "a" * 40,
            "WANDB_HTTP_TIMEOUT": "1"
        }
        res = sh.python("train.py", epochs=10, _bg=True, _env=environ)
        stdout, stderr = "", ""
        try:
            time.sleep(10)  # TODO: this might not always be enough...
            res.kill()
            res.stdout
        except (sh.ErrorReturnCode, ProcessLookupError):  # noqa: F821
            pass
        stdout = res.stdout.decode("utf8")
        stderr = res.stderr.decode("utf8")
        print(stdout)
        print(stderr)
        print("-----")
        if os.path.exists("wandb/debug.log"):
            print(open("wandb/debug.log").read())
        assert "Finished" in stdout
        assert "wandb: Network error" in stderr
        run_dir = glob.glob(os.path.join("wandb", "run*"))[0]
        meta = json.loads(
            open(os.path.join(run_dir, "wandb-metadata.json")).read())
        assert meta["state"] == "finished"


@pytest.mark.skipif(sys.version_info < (3, 6), reason="Funky things in python 2 land and multiprocessing")
def test_mock_server_with_internet(runner, live_mock_server):
    with runner.isolated_filesystem():
        with open("train.py", "w") as f:
            f.write(train_py)
        environ = {
            "WANDB_BASE_URL": "http://localhost:%i" % 8765,
            "WANDB_API_KEY": "a" * 40,
            "WANDB_HTTP_TIMEOUT": "1"
        }

        res = sh.python("train.py", epochs=10, _bg=True, _env=environ)
        stdout, stderr = "", ""
        try:
            res.wait()
        except (sh.ErrorReturnCode, ProcessLookupError):  # noqa: F821
            pass
        print(res)
        stdout = res.stdout.decode("utf8")
        stderr = res.stderr.decode("utf8")
        if os.path.exists("wandb/debug.log"):
            print(open("wandb/debug.log").read())
        assert "Finished" in stdout
        assert "wandb: Network error" not in stderr
        assert "wandb: Program ended successfully" in stderr
        run_dir = glob.glob(os.path.join("wandb", "run*"))[0]
        meta = json.loads(
            open(os.path.join(run_dir, "wandb-metadata.json")).read())
        assert meta["state"] == "finished"
